#include <os.h>
#include "disp.h"

#define FLOORY 40
#define FLOORTHICKNESS 8

#define RUNNERHEIGHT 20
#define RUNNERWIDTH 14
#define RUNNERX 40

#define GRAVESTONEHEIGHT 12
#define GRAVESTONEWIDTH 14

#define BLOCKHEIGHTWIDTH 8
#define BLOCKDISTANCE 2

#define BLOCKGROUPNUM 4

uint8_t blockgroups[BLOCKGROUPNUM][2] = {{0x00,0x00},{0x1C,0x00},{0x00,0x08},{0x00,0x18}};//blockgroupformat: 8 blockmgl, 2 bits pro x-position -> jede blockgruppe braucht fr obere reihe ein byte und fr untere eins, oberes [][0] unteres [][1]

uint8_t linerunner[140] = {
    0xFF,0xFF,0xFF,0xFF,0x00,0xFF,0xFF,
    0xFF,0xFF,0xFF,0xF0,0x00,0x0F,0xFF,
    0xFF,0xFF,0xFF,0xF0,0x00,0x0F,0xFF,
    0xFF,0xFF,0xFF,0xF0,0x00,0x0F,0xFF,
    0xFF,0xFF,0xFF,0xFF,0x00,0xFF,0xFF,
    0xFF,0xFF,0xFF,0xFF,0x00,0xFF,0xFF,
    0xFF,0xFF,0xFF,0x00,0x00,0xFF,0xFF,
    0xFF,0xFF,0xF0,0x00,0x00,0xFF,0xFF,
    0xFF,0xFF,0x70,0xF0,0x0F,0x01,0xFF,
    0xFF,0xFF,0x00,0xF0,0x0F,0xF0,0x07,
    0xFF,0xFF,0xF0,0xF0,0x0F,0xFF,0xFF,
    0xFF,0xFF,0xFF,0xF0,0x0F,0xFF,0xFF,
    0xFF,0xFF,0xFF,0xF0,0x07,0xFF,0xFF,
    0xFF,0xFF,0xF7,0x0B,0x70,0xFF,0xFF,
    0xFF,0xFF,0x70,0x0F,0xFF,0x0F,0xFF,
    0xFF,0xFF,0x00,0xFF,0xFF,0x70,0xFF,
    0x00,0x00,0xFF,0xFF,0xFF,0xF0,0x0F,
    0x07,0xBF,0xFF,0xFF,0xFF,0xF7,0x07,
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xF0,
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFB
};

uint8_t linerunner_rolling[32] = {
    0xFF,0xF0,0x0F,0xFF,
    0xF0,0x00,0x00,0x0F,
    0xF0,0x00,0x00,0x0F,
    0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,
    0xF0,0x00,0x00,0x0F,
    0xF0,0x00,0x00,0x0F,
    0xFF,0xF0,0x0F,0xFF
};

uint8_t dead[84] = {
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
    0xFF,0xFF,0xF0,0x00,0x0F,0xFF,0xFF,
    0xFF,0xFF,0x00,0xBB,0x00,0xFF,0xFF,
    0xFF,0xF0,0x0B,0x00,0xB0,0x0F,0xFF,
    0xFF,0xF0,0xBB,0x00,0xBB,0x0F,0xFF,
    0xFF,0x00,0xB0,0x00,0x0B,0x00,0xFF,
    0xFF,0x0B,0xB0,0x00,0x0B,0xB0,0xFF,
    0xF0,0x0B,0xBB,0x00,0xBB,0xB0,0x0F,
    0xF0,0xBB,0xBB,0x00,0xBB,0xBB,0x0F,
    0x00,0xBB,0xBB,0x00,0xBB,0xBB,0x00,
    0x0B,0xBB,0xBB,0xBB,0xBB,0xBB,0xB0,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00
};

inline void drawfloor(uint8_t* scrbuf){
    memset(scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS)*SCREEN_WIDTH/2, 0x00, FLOORTHICKNESS*SCREEN_WIDTH/2);
}

int drawrunner(uint8_t* scrbuf, signed char jumpheight){
    uint8_t* bufaddr;
	uint8_t* spraddr;
	int row, doublecol;
	int retval=0;
	if(jumpheight>=0){
	    bufaddr = scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-RUNNERHEIGHT-jumpheight)*SCREEN_WIDTH/2 + RUNNERX/2;
	    spraddr = linerunner;
        for(row=0;row<RUNNERHEIGHT;row++){
            for(doublecol=0;doublecol<RUNNERWIDTH/2;doublecol++){
                if(!(*bufaddr) && (*spraddr!=0xFF))//if runner touches black pixel
                    retval=1;
                if(*spraddr != 0xFF)
                    *bufaddr = *spraddr;
                spraddr++;
                bufaddr++;
            }
            bufaddr += (SCREEN_WIDTH-RUNNERWIDTH)/2;
        }
	}
	else{
	    bufaddr = scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH)*SCREEN_WIDTH/2 + RUNNERX/2;
	    spraddr = linerunner_rolling;
	    for(row=0;row<BLOCKHEIGHTWIDTH;row++){
            for(doublecol=0;doublecol<BLOCKHEIGHTWIDTH/2;doublecol++){
                if(!(*bufaddr) && (*spraddr!=0xFF))//if runner touches black pixel
                    retval=1;
                if(*spraddr != 0xFF)
                    *bufaddr = *spraddr;
                spraddr++;
                bufaddr++;
            }
            bufaddr += (SCREEN_WIDTH-BLOCKHEIGHTWIDTH)/2;
        }
	}
	return retval;
}

void drawgameover(uint8_t* scrbuf){
    uint8_t* bufaddr = scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-GRAVESTONEHEIGHT)*SCREEN_WIDTH/2 + RUNNERX/2;
	uint8_t* spraddr = dead;
	int row, doublecol;
	for(row=0;row<GRAVESTONEHEIGHT;row++){
	    for(doublecol=0;doublecol<GRAVESTONEWIDTH/2;doublecol++){
            if(*spraddr != 0xFF)
                *bufaddr = *spraddr;
            spraddr++;
            bufaddr++;
	    }
	    bufaddr += (SCREEN_WIDTH-GRAVESTONEWIDTH)/2;
	}
}

void drawlowblock(uint8_t* scrbuf, short x, char offset){
    short i;
    if(x-offset>=SCREEN_WIDTH-BLOCKHEIGHTWIDTH){
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            memset(scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH+i)*SCREEN_WIDTH/2 + (x-offset)/2, 0x00, offset/2);
    }
    if(x>offset){
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            memset(scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH+i)*SCREEN_WIDTH/2 + (x-offset)/2, 0x00, BLOCKHEIGHTWIDTH/2);
    }
    else{
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            memset(scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-BLOCKHEIGHTWIDTH+i)*SCREEN_WIDTH/2, 0x00, (BLOCKHEIGHTWIDTH+x-offset)/2);
    }
}

void drawhighblock(uint8_t* scrbuf, short x, char offset){
    short i;
    if(x-offset>=SCREEN_WIDTH-BLOCKHEIGHTWIDTH){
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            memset(scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-3*BLOCKHEIGHTWIDTH+i)*SCREEN_WIDTH/2 + (x-offset)/2, 0x00, offset/2);
    }
    if(x>offset){
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            memset(scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-3*BLOCKHEIGHTWIDTH+i)*SCREEN_WIDTH/2 + (x-offset)/2, 0x00, BLOCKHEIGHTWIDTH/2);
    }
    else{
        for(i=0;i<BLOCKHEIGHTWIDTH;i++)
            memset(scrbuf + (SCREEN_HEIGHT-FLOORY-FLOORTHICKNESS-3*BLOCKHEIGHTWIDTH+i)*SCREEN_WIDTH/2, 0x00, (BLOCKHEIGHTWIDTH+x-offset)/2);
    }
}

void drawallblocks(uint8_t* scrbuf, uint64_t currentblockgroups[2], char offset){
    int i;
    //low blocks:
    if(currentblockgroups[0] & 0xFF00000000000000){//if on the left side low blocks
        for(i=0;i<8;i++){
            if(currentblockgroups[0] & ((uint64_t)0x1 << (63-i)))//cast needed because shift only works for the bigness of the type
                drawlowblock(scrbuf, i*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
        }
    }
    if(currentblockgroups[0] & 0x00FF000000000000){
        for(i=0;i<8;i++){
            if(currentblockgroups[0] & ((uint64_t)0x1 << (55-i)))//cast needed because shift only works for the bigness of the type
                drawlowblock(scrbuf, (8+i)*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
        }
    }
    if(currentblockgroups[0] & 0x0000FF0000000000){
        for(i=0;i<8;i++){
            if(currentblockgroups[0] & ((uint64_t)0x1 << (47-i)))//cast needed because shift only works for the bigness of the type
                drawlowblock(scrbuf, (16+i)*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
        }
    }
    if(currentblockgroups[0] & 0x000000FF00000000){
        for(i=0;i<8;i++){
            if(currentblockgroups[0] & ((uint64_t)0x1 << (39-i)))//cast needed because shift only works for the bigness of the type
                drawlowblock(scrbuf, (24+i)*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
        }
    }
    //if((currentblockgroups[0] & 0x0000000080000000) && offset)
    //    drawlowblock(scrbuf, 32*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
    //high blocks:
    if(currentblockgroups[1] & 0xFF00000000000000){//if on the left side high blocks
        for(i=0;i<8;i++){
            if(currentblockgroups[1] & ((uint64_t)0x1 << (63-i)))//cast needed because shift only works for the bigness of the type
                drawhighblock(scrbuf, i*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
        }
    }
    if(currentblockgroups[1] & 0x00FF000000000000){
        for(i=0;i<8;i++){
            if(currentblockgroups[1] & ((uint64_t)0x1 << (55-i)))//cast needed because shift only works for the bigness of the type
                drawhighblock(scrbuf, (8+i)*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
        }
    }
    if(currentblockgroups[1] & 0x0000FF0000000000){
        for(i=0;i<8;i++){
            if(currentblockgroups[1] & ((uint64_t)0x1 << (47-i)))//cast needed because shift only works for the bigness of the type
                drawhighblock(scrbuf, (16+i)*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
        }
    }
    if(currentblockgroups[1] & 0x000000FF00000000){
        for(i=0;i<8;i++){
            if(currentblockgroups[1] & ((uint64_t)0x1 << (39-i)))//cast needed because shift only works for the bigness of the type
                drawhighblock(scrbuf, (24+i)*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
        }
    }
    //if((currentblockgroups[1] & 0x0000000080000000) && offset)
    //    drawhighblock(scrbuf, 32*(BLOCKHEIGHTWIDTH + BLOCKDISTANCE), offset);
}

void reload(uint64_t currentblockgroups[2]){
    char group;
    group = *((uint16_t*)0x900D000C) % BLOCKGROUPNUM;//second timer 2 modulo the number of blockgroups
    currentblockgroups[0] += blockgroups[group][0];
    currentblockgroups[1] += blockgroups[group][1];
}

int main(){
    clrscr();
    lcd_ingray();
    uint8_t* scrbuf = (uint8_t*) malloc(SCREEN_BYTES_SIZE);
    signed char jumpheight=0, canjump = 1, rolling = 0, canroll=0;
    uint64_t currentblockgroups[2];//least significant byte first -> little-endian
    char gameover=0;
    char offset=0;
    char loadgroups = 0;
    int i;
    unsigned restore_cpu_speed = set_cpu_speed(CPU_SPEED_90MHZ);
    currentblockgroups[0] = 0x0000000000000000 + (blockgroups[1][0] * 4294967296) + (blockgroups[2][0] * 256);
    currentblockgroups[1] = 0x0000000000000000 + (blockgroups[1][1] * 4294967296) + (blockgroups[2][1] * 256);
    reload(currentblockgroups);//load first group
    while(!isKeyPressed(KEY_NSPIRE_ESC)){
        sleep(20);
        memset(scrbuf, 0xFF, SCREEN_BYTES_SIZE);
        drawfloor(scrbuf);
        drawallblocks(scrbuf, currentblockgroups, offset);
        gameover = drawrunner(scrbuf, jumpheight);
        memcpy(SCREEN_BASE_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);
        offset+=2;
        if(offset == BLOCKHEIGHTWIDTH+BLOCKDISTANCE){
            offset=0;
            currentblockgroups[0]<<=1;
            currentblockgroups[1]<<=1;
            loadgroups++;
        }
        if(loadgroups == 8){//if only 56 bits/blocks left, reload to 64
            reload(currentblockgroups);
            loadgroups = 0;
        }
        if(isKeyPressed(KEY_NSPIRE_LEFT)){
            if(jumpheight <= 0)
                canjump = 1;
            if(jumpheight<0)
                jumpheight=0;
            if((jumpheight<4*BLOCKHEIGHTWIDTH) && (canjump == 1))
                jumpheight+=BLOCKHEIGHTWIDTH;
            else//if cannot jump or at highest height
                canjump = 0;
        }
        else if(jumpheight>0){
            canjump = 0;
        }
        if(isKeyPressed(KEY_NSPIRE_RIGHT) && (canroll == 1)){
            jumpheight = -1;
            canroll = 0;
            rolling = 5*BLOCKHEIGHTWIDTH;
        }
        if(jumpheight == -1){
            rolling -= 2;
            if(rolling <= 0){
                jumpheight=0;
                canroll = 0;
            }
        }
        else{
            canroll = 1;
        }
        if(jumpheight>0){
            if(jumpheight>1)
                jumpheight -= 1;
            else
                jumpheight = 0;
        }
        if(gameover){
            drawgameover(scrbuf);
            memcpy(SCREEN_BASE_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);
            wait_no_key_pressed();
            wait_key_pressed();
            break;
        }
    }
    set_cpu_speed(restore_cpu_speed);
    free(scrbuf);
    return 0;
}
